#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "ReadOnlyRun",
    "get_attribute_value_from_entry",
]

from typing import (
    TYPE_CHECKING,
    Generator,
    List,
    Optional,
    Union,
)

from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import QualifiedName
from neptune.internal.utils import verify_type
from neptune.metadata_containers.metadata_containers_table import TableEntry

from neptune_fetcher.fetchable import (
    Downloadable,
    Fetchable,
    FetchableSeries,
    which_fetchable,
)

if TYPE_CHECKING:
    from neptune_fetcher.read_only_project import ReadOnlyProject


def get_attribute_value_from_entry(entry: TableEntry, name: str) -> Optional[str]:
    try:
        return entry.get_attribute_value(name)
    except ValueError:
        return None


class ReadOnlyRun:
    def __init__(self, read_only_project: "ReadOnlyProject", with_id: str) -> None:
        self.project = read_only_project
        self.with_id = with_id

        verify_type("with_id", with_id, str)

        self._container_id = QualifiedName(f"{self.project.project_identifier}/{with_id}")
        self._cache = dict()
        self._structure = {
            attribute.path: which_fetchable(
                attribute,
                self.project._backend,
                self._container_id,
                self._cache,
            )
            for attribute in self.project._backend.get_attributes(self._container_id, ContainerType.RUN)
        }

    def __getitem__(self, item: str) -> Union[Fetchable, FetchableSeries, Downloadable]:
        return self._structure[item]

    def __delitem__(self, key: str) -> None:
        del self._cache[key]

    @property
    def field_names(self) -> Generator[str, None, None]:
        """Lists names of run fields.

        Returns a generator of run fields.
        """
        yield from self._structure

    def prefetch(self, paths: List[str]) -> None:
        """Prefetches values of a list of fields and stores them in local cache.

        Args:
            paths: List of field paths to prefetch.
        """
        fetched = self.project._backend.prefetch_values(self._container_id, ContainerType.RUN, paths)
        self._cache.update(fetched)



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "initialize",
]

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
)

from neptune_experimental.utils import wrap_method

if TYPE_CHECKING:
    from neptune.internal.operation_processors.async_operation_processor import AsyncOperationProcessor


def initialize() -> None:
    from neptune.internal.operation_processors.async_operation_processor import AsyncOperationProcessor

    wrap_method(obj=AsyncOperationProcessor, method="__init__", wrapper=custom_init_async_op_processor)
    wrap_method(obj=AsyncOperationProcessor, method="_check_queue_size", wrapper=custom_check_queue_size)
    wrap_method(obj=AsyncOperationProcessor.ConsumerThread, method="__init__", wrapper=custom_init_consumer_thread)
    wrap_method(obj=AsyncOperationProcessor.ConsumerThread, method="process_batch", wrapper=custom_process_batch)


class MonotonicIncBatchSize:
    def __init__(self, size_limit: int, initial_size: int = 10, scale_coef: float = 1.6):
        if size_limit <= 0 or not isinstance(size_limit, int):
            raise ValueError("Size limit must be a positive integer")
        if scale_coef <= 1:
            raise ValueError("Scale coefficient cannot be smaller than 1")
        if initial_size <= 0 or initial_size > size_limit or not isinstance(initial_size, int):
            raise ValueError(f"Initial size must be an integer in the interval (0, {size_limit}>")

        self._size_limit: int = size_limit
        self._current_size: int = initial_size
        self._scale_coef: float = scale_coef

    def increase(self) -> None:
        self._current_size = min(int(self._current_size * self._scale_coef + 1), self._size_limit)

    def get(self) -> int:
        return self._current_size


def custom_init_async_op_processor(
    self: "AsyncOperationProcessor", *args: Any, original: Callable[..., Any], batch_size: int = 1000, **kwargs: Any
) -> None:
    self._incremental_batch_size = MonotonicIncBatchSize(size_limit=batch_size)
    original(self, *args, batch_size=batch_size, **kwargs)


def custom_check_queue_size(
    self: "AsyncOperationProcessor", *args: Any, original: Callable[..., bool], **kwargs: Any
) -> bool:
    if hasattr(self, "_incremental_batch_size") and self._incremental_batch_size is not None:
        return bool(self._queue.size() > self._incremental_batch_size.get() / 2)
    return original(*args, **kwargs)


def custom_init_consumer_thread(
    self: "AsyncOperationProcessor.ConsumerThread", *args: Any, original: Callable[..., Any], **kwargs: Any
) -> None:
    original(self, *args, **kwargs)
    self._batch_size = self._processor._incremental_batch_size.get()


def custom_process_batch(
    self: "AsyncOperationProcessor.ConsumerThread", *args: Any, original: Callable[..., Any], **kwargs: Any
) -> None:
    original(self, *args, **kwargs)
    if self._processor._incremental_batch_size is not None:
        self._processor._incremental_batch_size.increase()
        self._batch_size = self._processor._incremental_batch_size.get()



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import contextlib
import os
import threading
from pathlib import Path
from queue import Queue
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

from neptune.constants import (
    ASYNC_DIRECTORY,
    NEPTUNE_DATA_DIRECTORY,
)
from neptune.core.components.abstract import WithResources
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import UniqueId
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.async_operation_processor import AsyncOperationProcessor
from neptune.internal.operation_processors.operation_logger import (
    ProcessorStopLogger,
    ProcessorStopSignalType,
)
from neptune.internal.operation_processors.operation_processor import OperationProcessor
from neptune.internal.operation_processors.utils import get_container_dir
from neptune.internal.utils.logger import get_logger

if TYPE_CHECKING:
    from neptune.core.components.abstract import Resource
    from neptune.core.components.operation_storage import OperationStorage
    from neptune.internal.operation_processors.operation_logger import (
        ProcessorStopSignal,
        ProcessorStopSignalData,
    )
    from neptune.internal.signals_processing.signals import Signal


_logger = get_logger()

SIGNALS_TO_ACCUMULATE = (
    ProcessorStopSignalType.SUCCESS,
    ProcessorStopSignalType.WAITING_FOR_OPERATIONS,
    ProcessorStopSignalType.STILL_WAITING,
)


def get_container_partition_path(type_dir: str, execution_dir_name: str, partition: int) -> Path:
    neptune_data_dir = Path(os.getenv("NEPTUNE_DATA_DIRECTORY", NEPTUNE_DATA_DIRECTORY))
    return neptune_data_dir / type_dir / (execution_dir_name + "__partition_" + str(partition))


def get_container_path(type_dir: str, execution_dir_name: str) -> Path:
    neptune_data_dir = Path(os.getenv("NEPTUNE_DATA_DIRECTORY", NEPTUNE_DATA_DIRECTORY))
    return neptune_data_dir / type_dir / execution_dir_name


class ProcessorStopSignalHandler:
    def __init__(self, processor_stop_logger: ProcessorStopLogger, num_processors: int) -> None:
        self._logger = processor_stop_logger
        self._active_processors = num_processors
        self._total_op_count = 0
        self._still_waiting_signals: Dict[int, int] = {}

    def handle_connection_interruption(self, signal: "ProcessorStopSignal") -> None:
        self._logger.log_connection_interruption(signal.data.max_reconnect_wait_time)

    def handle_waiting_for_operations(
        self,
        signals: Dict[ProcessorStopSignalType, "List[ProcessorStopSignalData]"],
    ) -> None:
        self._total_op_count = sum(
            sig_data.size_remaining for sig_data in signals[ProcessorStopSignalType.WAITING_FOR_OPERATIONS]
        )
        self._logger.log_remaining_operations(size_remaining=self._total_op_count)

        # reset the array to wait for the next 'batch' of wait signals
        signals[ProcessorStopSignalType.WAITING_FOR_OPERATIONS] = []

    def handle_failure(
        self,
        signal: "ProcessorStopSignal",
    ) -> None:
        # log sync failure
        if signal.signal_type == ProcessorStopSignalType.SYNC_FAILURE:
            self._logger.log_sync_failure(signal.data.seconds, signal.data.size_remaining)

        # log reconnect failure
        elif signal.signal_type == ProcessorStopSignalType.RECONNECT_FAILURE:
            self._logger.log_reconnect_failure(signal.data.max_reconnect_wait_time, signal.data.size_remaining)

    def handle_success(
        self,
        signals: Dict[ProcessorStopSignalType, "List[ProcessorStopSignalData]"],
    ) -> None:
        ops_synced = sum(sig_data.already_synced for sig_data in signals[ProcessorStopSignalType.SUCCESS])
        self._logger.log_success(ops_synced=ops_synced)

    def mark_processor_as_completed(self, signal: "ProcessorStopSignal") -> None:
        self._still_waiting_signals.pop(signal.data.processor_id, None)
        self._active_processors -= 1

    def handle_still_waiting(
        self,
        signals: Dict[ProcessorStopSignalType, "List[ProcessorStopSignalData]"],
        waiting_signal: "ProcessorStopSignal",
    ) -> None:
        assert waiting_signal.signal_type == ProcessorStopSignalType.STILL_WAITING

        self._still_waiting_signals[waiting_signal.data.processor_id] = waiting_signal.data.already_synced

        # wait for all active processors to send their still waiting signals
        if len(signals[ProcessorStopSignalType.STILL_WAITING]) != self._active_processors:
            return

        # reset the array to wait for the next 'batch' of still waiting signals
        signals[ProcessorStopSignalType.STILL_WAITING] = []

        synced_by_success_proc = sum(sig_data.already_synced for sig_data in signals[ProcessorStopSignalType.SUCCESS])
        synced_by_waiting_proc = sum([already_synced for already_synced in self._still_waiting_signals.values()])

        total_size_synced = synced_by_success_proc + synced_by_waiting_proc

        total_size_remaining = self._total_op_count - total_size_synced

        self._logger.log_still_waiting(
            size_remaining=total_size_remaining,
            already_synced=total_size_synced,
            already_synced_proc=total_size_synced * 100 / self._total_op_count,
        )


class ProcessorStopEventListener(contextlib.AbstractContextManager):
    def __init__(self, num_processors: int, signal_queue: "Queue[ProcessorStopSignal]") -> None:
        self._num_processors = num_processors
        self._signal_queue = signal_queue
        self._t = threading.Thread(target=self.work)
        self._logger = ProcessorStopLogger(processor_id=0, signal_queue=None, logger=_logger)

    def __enter__(self) -> "ProcessorStopEventListener":
        self._t.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if exc_val is not None:
            _logger.error(exc_val, exc_info=exc_tb)
        self._t.join()

    def work(self) -> None:
        signals: Dict[ProcessorStopSignalType, "List[ProcessorStopSignalData]"] = {
            signal_type: [] for signal_type in SIGNALS_TO_ACCUMULATE
        }

        handler = ProcessorStopSignalHandler(self._logger, self._num_processors)

        while True:  # only a single failure or full success breaks the loop
            try:
                signal = self._signal_queue.get()

                if signal.signal_type in SIGNALS_TO_ACCUMULATE:
                    signals[signal.signal_type].append(signal.data)

                # log connection interruption
                if signal.signal_type == ProcessorStopSignalType.CONNECTION_INTERRUPTED:
                    handler.handle_connection_interruption(signal=signal)

                # log that we wait for operations
                if len(signals[ProcessorStopSignalType.WAITING_FOR_OPERATIONS]) == self._num_processors:
                    handler.handle_waiting_for_operations(signals=signals)

                # if something went wrong - log failure
                if signal.signal_type in (
                    ProcessorStopSignalType.SYNC_FAILURE,
                    ProcessorStopSignalType.RECONNECT_FAILURE,
                ):
                    handler.handle_failure(signal=signal)
                    return

                # mark processor as completed - don't count its operations in the 'still waiting' count
                if signal.signal_type == ProcessorStopSignalType.SUCCESS:
                    handler.mark_processor_as_completed(signal=signal)

                # if all went well - log success
                if len(signals[ProcessorStopSignalType.SUCCESS]) == self._num_processors:
                    handler.handle_success(signals=signals)
                    return

                # log that we still wait + percentage of already synced operations
                if signal.signal_type == ProcessorStopSignalType.STILL_WAITING:
                    handler.handle_still_waiting(signals=signals, waiting_signal=signal)
            except KeyboardInterrupt:
                return


class PartitionedOperationProcessor(WithResources, OperationProcessor):
    def __init__(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        backend: NeptuneBackend,
        lock: threading.RLock,
        queue: "Queue[Signal]",
        batch_size: int,
        sleep_time: float = 5,
        partitions: int = 5,
    ):
        self._execution_dir_name = get_container_dir(container_id, container_type)
        self._data_path = get_container_path(ASYNC_DIRECTORY, get_container_dir(container_id, container_type))
        self._partitions = partitions
        self._processors = [
            AsyncOperationProcessor(
                container_id=container_id,
                container_type=container_type,
                backend=backend,
                lock=lock,
                queue=queue,
                sleep_time=sleep_time,
                batch_size=batch_size,
                data_path=get_container_partition_path(ASYNC_DIRECTORY, self._execution_dir_name, partition_id),
                should_print_logs=False,
            )
            for partition_id in range(partitions)
        ]

    def enqueue_operation(self, op: Operation, *, wait: bool) -> None:
        processor = self._get_operation_processor(op.path)
        processor.enqueue_operation(op, wait=wait)

    @property
    def operation_storage(self) -> "OperationStorage":
        # This is a bit of a hack as we assume that all processors use the same operation storage
        #   this could make problems when the first processor will be cleaned up earlier than the others
        return self._processors[0].operation_storage

    @property
    def data_path(self) -> Path:
        # This is a bit of a hack as we assume that all processors use the same data path
        #   this could make problems when the first processor will be cleaned up earlier than the others
        # mypy had a problem with the return type of the property, so I had to cast it to Path
        return Path(self._processors[0].data_path)

    @property
    def resources(self) -> Tuple["Resource", ...]:
        return tuple(self._processors)

    def _get_operation_processor(self, path: List[str]) -> OperationProcessor:
        path_hash = hash(tuple(path))
        return self._processors[path_hash % self._partitions]

    def pause(self) -> None:
        for processor in self._processors:
            processor.pause()

    def resume(self) -> None:
        for processor in self._processors:
            processor.resume()

    def wait(self) -> None:
        for processor in self._processors:
            processor.wait()

    def start(self) -> None:
        # TODO: Handle exceptions
        for processor in self._processors:
            processor.start()

    def stop(self, seconds: Optional[float] = None) -> None:
        # TODO: Handle exceptions

        signal_queue: "Queue[ProcessorStopSignal]" = Queue()

        threads = [
            threading.Thread(target=processor.stop, args=(seconds, signal_queue)) for processor in self._processors
        ]

        with ProcessorStopEventListener(num_processors=len(self._processors), signal_queue=signal_queue):
            for t in threads:
                t.start()

            for t in threads:
                t.join()



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["CustomBackend"]

from typing import (
    Any,
    Dict,
    List,
)

from bravado.exception import HTTPNotFound
from neptune.exceptions import ContainerUUIDNotFound
from neptune.internal.backends.api_model import (
    Attribute,
    AttributeType,
)
from neptune.internal.backends.hosted_client import DEFAULT_REQUEST_KWARGS
from neptune.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.utils.logger import get_logger

from neptune_fetcher.attributes import (
    Attr,
    Boolean,
    Datetime,
    Float,
    Integer,
    String,
)

logger = get_logger()


def to_attribute(attr) -> Attribute:
    return Attribute(attr.name, AttributeType(attr.type))


def get_attribute_from_dto(dto: Any) -> Attr:
    if dto.stringProperties is not None:
        return String(AttributeType(dto.type), dto.stringProperties.value)
    if dto.boolProperties is not None:
        return Boolean(AttributeType(dto.type), dto.boolProperties.value)
    if dto.floatProperties is not None:
        return Float(AttributeType(dto.type), dto.floatProperties.value)
    if dto.intProperties is not None:
        return Integer(AttributeType(dto.type), dto.intProperties.value)
    if dto.datetimeProperties is not None:
        return Datetime(AttributeType(dto.type), dto.datetimeProperties.value)
    raise Exception(f"Attribute {dto.name} of type {AttributeType(dto.type)} does not support prefetching")


class CustomBackend(HostedNeptuneBackend):
    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[Attribute]:
        params = {
            "experimentIdentifier": container_id,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            experiment = self.leaderboard_client.api.queryAttributeDefinitions(**params).response().result

            attribute_type_names = [at.value for at in AttributeType]
            accepted_attributes = [attr for attr in experiment.entries if attr.type in attribute_type_names]

            # Notify about ignored attrs
            ignored_attributes = set(attr.type for attr in experiment.entries) - set(
                attr.type for attr in accepted_attributes
            )
            if ignored_attributes:
                logger.warning(
                    "Ignored following attributes (unknown type): %s.\n" "Try to upgrade `neptune`.",
                    ignored_attributes,
                )

            return [to_attribute(attr) for attr in accepted_attributes if attr.type in attribute_type_names]
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(
                container_id=container_id,
                container_type=container_type,
            ) from e

    def prefetch_values(self, container_id: str, container_type: ContainerType, paths: List[str]) -> Dict[str, Attr]:
        params = {
            "holderIdentifier": container_id,
            "holderType": "experiment",
            "attributeQuery": {
                "attributePathsFilter": paths,
            },
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            result = self.leaderboard_client.api.getAttributesWithPathsFilter(**params).response().result
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(
                container_id=container_id,
                container_type=container_type,
            ) from e

        return {dto.name: get_attribute_from_dto(dto) for dto in result.attributes}



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["wrap_method", "override_method"]

from functools import wraps
from typing import (
    Any,
    Callable,
)


def wrap_method(*, obj: Any, method: str, wrapper: Callable[..., Any]) -> None:
    def wrapped_method_factory(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def new_method(*args: Any, **kwargs: Any) -> Any:
            return wrapper(*args, original=func, **kwargs)

        return new_method

    override_method(obj=obj, method=method, method_factory=wrapped_method_factory)


def override_method(
    *, obj: Any, method: str, method_factory: Callable[[Callable[..., Any]], Callable[..., Any]]
) -> None:
    source = getattr(obj, method)
    new_method = method_factory(source)
    setattr(obj, method, new_method)



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "ReadOnlyProject",
]

import os
from typing import (
    TYPE_CHECKING,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
)

from neptune import Project
from neptune.envs import PROJECT_ENV_NAME
from neptune.internal.backends.nql import NQLEmptyQuery
from neptune.internal.backends.project_name_lookup import project_name_lookup
from neptune.internal.container_type import ContainerType
from neptune.internal.credentials import Credentials
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
    conform_optional,
)
from neptune.management.internal.utils import normalize_project_name
from neptune.metadata_containers.metadata_containers_table import Table
from neptune.metadata_containers.utils import prepare_nql_query
from neptune.typing import ProgressBarType

from neptune_fetcher.custom_backend import CustomBackend
from neptune_fetcher.read_only_run import (
    ReadOnlyRun,
    get_attribute_value_from_entry,
)

if TYPE_CHECKING:
    from pandas import DataFrame


class ReadOnlyProject:
    """Class for retrieving metadata from a neptune.ai project in a limited read-only mode."""

    def __init__(
        self,
        project: Optional[str] = None,
        api_token: Optional[str] = None,
        proxies: Optional[dict] = None,
    ) -> None:
        """Initializes a Neptune project in limited read-only mode.

        Compared to a regular Project object, it contains only basic project information and exposes more lightweight
        methods for fetching run metadata.

        Args:
            project: The name of the Neptune project.
            api_token: Neptune account's API token.
                If left empty, the value of the NEPTUNE_API_TOKEN environment variable is used (recommended).
            proxies: A dictionary of proxy settings if needed.
        """
        self._project: Optional[str] = project if project else os.getenv(PROJECT_ENV_NAME)
        if self._project is None:
            raise Exception("Could not find project name in env")

        self._backend: CustomBackend = CustomBackend(
            credentials=Credentials.from_token(api_token=api_token), proxies=proxies
        )
        self._project_qualified_name: Optional[str] = conform_optional(self._project, QualifiedName)
        self._project_api_object: Project = project_name_lookup(
            backend=self._backend, name=self._project_qualified_name
        )
        self._project_id: UniqueId = self._project_api_object.id
        self.project_identifier = normalize_project_name(
            name=self._project, workspace=self._project_api_object.workspace
        )

    def list_runs(self) -> Generator[Dict[str, Optional[str]], None, None]:
        """Lists IDs and names of the runs in the project.

        Returns a generator of run info dictionaries `{"sys/id": ..., "sys/name": ...}`.
        """
        leaderboard_entries = self._backend.search_leaderboard_entries(
            project_id=self._project_id,
            types=[ContainerType.RUN],
            query=NQLEmptyQuery(),
            columns=["sys/id", "sys/name"],
        )

        for row in Table(
            backend=self._backend, container_type=ContainerType.RUN, entries=leaderboard_entries
        ).to_rows():
            yield {
                "sys/id": get_attribute_value_from_entry(entry=row, name="sys/id"),
                "sys/name": get_attribute_value_from_entry(entry=row, name="sys/name"),
            }

    def fetch_read_only_runs(self, with_ids: List[str]) -> Generator[ReadOnlyRun, None, None]:
        """Lists runs of the project in the form of read-only runs.

        Returns a generator of `ReadOnlyProject.ReadOnlyRun` instances.

        Args:
            with_ids: List of run ids to fetch.
        """
        for run_id in with_ids:
            yield ReadOnlyRun(read_only_project=self, with_id=run_id)

    def fetch_runs(self) -> "DataFrame":
        """Fetches a table containing IDs and names of runs in the project.

        Returns `pandas.DataFrame` with two columns ('sys/id' and 'sys/name') and rows corresponding to project runs.
        """
        return self.fetch_runs_df(columns=["sys/id", "sys/name"])

    def fetch_runs_df(
        self,
        columns: Optional[Iterable[str]] = None,
        with_ids: Optional[Iterable[str]] = None,
        states: Optional[Iterable[str]] = None,
        owners: Optional[Iterable[str]] = None,
        tags: Optional[Iterable[str]] = None,
        trashed: Optional[bool] = False,
        limit: Optional[int] = None,
        sort_by: str = "sys/creation_time",
        ascending: bool = False,
        progress_bar: Optional[ProgressBarType] = None,
    ) -> "DataFrame":
        """Fetches the runs' metadata and returns them as a pandas DataFrame.

        Args:
            columns: A list of column names to include in the DataFrame.
                Defaults to None, which includes all available columns.
            with_ids: A list of run IDs to filter the results.
            states: A list of run states to filter the results.
            owners: A list of owner names to filter the results.
            tags: A list of tags to filter the results.
            trashed: Whether to return trashed runs as the result.
                If True: return only trashed runs.
                If False (default): return only non-trashed runs.
                If None: return all runs.
            limit: How many entries to return at most. If `None`, all entries are returned.
            sort_by: Name of the field to sort the results by.
                The field must represent a simple type (string, float, datetime, integer, or Boolean).
            ascending: Whether to sort the entries in ascending order of the sorting column values.
            progress_bar: Set to `False` to disable the download progress bar,
                or pass a `ProgressBarCallback` class to use your own progress bar callback.

        Returns:
            DataFrame: A pandas DataFrame containing information about the fetched runs.

        Example:
            ```
            # Fetch all runs with specific columns
            columns_to_fetch = ["sys/name", "sys/modification_time", "training/lr"]
            runs_df = my_project.fetch_runs_df(columns=columns_to_fetch, states=["active"])

            # Fetch runs by specific IDs
            specific_run_ids = ["RUN-123", "RUN-456"]
            specific_runs_df = my_project.fetch_runs_df(with_ids=specific_run_ids)
            ```
        """
        query = prepare_nql_query(with_ids, states, owners, tags, trashed)

        if columns is not None:
            # always return entries with `sys/id` column when filter applied
            columns = set(columns)
            columns.add("sys/id")

        leaderboard_entries = self._backend.search_leaderboard_entries(
            project_id=self._project_id,
            types=[ContainerType.RUN],
            query=query,
            columns=columns,
            limit=limit,
            sort_by=sort_by,
            ascending=ascending,
            progress_bar=progress_bar,
        )

        return Table(
            backend=self._backend,
            container_type=ContainerType.RUN,
            entries=leaderboard_entries,
        ).to_pandas()
